/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.control.activity.scheduler;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.manager.EndOfTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.binding.ContextualBindingController;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Scheduler for periodic task. Executes each task at a given period. This
 * scheduler must be executed by only one thread.
 */
public class PeriodicSchedulerImpl
    implements
      Scheduler,
      InterruptTask,
      ContextualBindingController
{

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected Map<String, Object> tasksItf = new HashMap<String, Object>();
  protected EndOfTask           endOfTaskItf;

  // ---------------------------------------------------------------------------
  // Fields
  // ---------------------------------------------------------------------------

  protected Logger              logger;
  protected Thread              executingThread;
  protected PeriodicTask        executingTask;
  protected boolean             interrupted;
  protected PeriodicTask        tasksQueue;
  protected Object              lock     = new Object();

  // ---------------------------------------------------------------------------
  // Implementation of the Scheduler interface
  // ---------------------------------------------------------------------------

  /**
   * @see Scheduler#schedule(int, Object)
   */
  public Object schedule(final int executionQuanta, final Object hints)
      throws InterruptedException
  {
    executingThread = Thread.currentThread();
    do
    {

      Object o = null;
      long currentTime = 0;
      try
      {
        synchronized (lock)
        {
          // wait for a task to schedule
          while (tasksQueue == null)
          {
            lock.wait();
          }

          executingTask = tasksQueue;

          // I've a task, wait for wake-up
          currentTime = System.currentTimeMillis();
          while (tasksQueue.wakeUpTime > currentTime)
          {
            lock.wait(tasksQueue.wakeUpTime - currentTime);
            currentTime = System.currentTimeMillis();
          }

          tasksQueue = tasksQueue.next;
        }
        o = executingTask.taskItf.execute(hints);
      }
      catch (InterruptedException e)
      {
        if (!interrupted)
        {
          // not my interruption !
          throw e;
        }
        logger.log(BasicLevel.DEBUG, "InterruptedException caught.");
      }
      catch (Throwable e)
      {
        logger.log(BasicLevel.INFO,
            "Exception caught during execution of a task: ", e);
        o = Task.STOP_EXECUTING;
      }
      synchronized (lock)
      {
        if (Task.STOP_EXECUTING.equals(o) || interrupted)
        {
          endOfTaskItf.taskEnded(executingTask.taskItf);
          executingTask = null;
          interrupted = false;
        }
        else
        {
          // reschedule task.
          long period = ((Number) o).longValue();
          if (period < 0)
          {
            endOfTaskItf.taskEnded(executingTask.taskItf);
          }
          else if (period == 0)
          {
            period = executingTask.period;
          }
          executingTask.wakeUpTime = currentTime + period;
          inserTask(executingTask);
          executingTask = null;
        }
      }
    }
    while (executionQuanta < 0);

    executingThread = null;
    return null;
  }

  /**
   * @see InterruptTask#interruptTask(Task)
   */
  public void interruptTask(Task task) throws IllegalTaskException
  {
    synchronized (lock)
    {
      if (executingTask != null && executingTask.taskItf == task)
      {
        // executing task
        interrupted = true;
        executingThread.interrupt();
      }
      else
      {
        // search task
        if (tasksQueue.taskItf == task)
        {
          // interrupted task is first one
          tasksQueue = tasksQueue.next;
          lock.notify();
        }
        else
        {
          PeriodicTask prev = tasksQueue;
          while (prev.next.taskItf != task)
          {
            prev = prev.next;
            if (prev.next == null)
            {
              throw new IllegalTaskException(task, "Unknown task");
            }
          }
          prev.next = prev.next.next;
        }
        endOfTaskItf.taskEnded(task);
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ContextualBindingController#bindFc(String, Object, Object)
   */
  public void bindFc(String clientItfName, Object serverItf, Object hints)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.startsWith(Task.ITF_NAME))
    {
      tasksItf.put(clientItfName, serverItf);
      long period = ((Number) hints).longValue();
      PeriodicTask task = new PeriodicTask(period, (Task) serverItf);
      inserTask(task);
    }
    else
    {
      bindFc(clientItfName, serverItf);
    }
  }

  /**
   * @see ContextualBindingController#bindFc(String, Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.startsWith(Task.ITF_NAME))
    {
      throw new IllegalBindingException("Requires a contextual binding");
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      endOfTaskItf = (EndOfTask) serverItf;
    }
    else if (clientItfName.equals("logger"))
    {
      logger = (Logger) serverItf;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see ContextualBindingController#lookupFc(String)
   */
  public Object lookupFc(String clientItfName) throws NoSuchInterfaceException
  {
    if (clientItfName.startsWith(Task.ITF_NAME))
    {
      Object itf = tasksItf.get(clientItfName);
      if (itf != null)
      {
        return itf;
      }
      throw new NoSuchInterfaceException(clientItfName);
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      return endOfTaskItf;
    }
    throw new NoSuchInterfaceException(clientItfName);
  }

  /**
   * @see ContextualBindingController#unbindFc(String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    if (clientItfName.startsWith(Task.ITF_NAME))
    {
      if (tasksItf.remove(clientItfName) == null)
      {
        throw new NoSuchInterfaceException(clientItfName);
      }
    }
    else if (clientItfName.equals(EndOfTask.ITF_NAME))
    {
      endOfTaskItf = null;
    }
    else if (clientItfName.equals("logger"))
    {
      logger = null;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see ContextualBindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Task.ITF_NAME, EndOfTask.ITF_NAME};
  }

  // ---------------------------------------------------------------------------
  // Utility method.
  // ---------------------------------------------------------------------------

  void inserTask(PeriodicTask task)
  {
    synchronized (lock)
    {
      if (tasksQueue == null || tasksQueue.compareTo(task) > 0)
      {
        // insert as first
        task.next = tasksQueue;
        tasksQueue = task;
        lock.notify();
      }
      else
      {
        PeriodicTask prev = tasksQueue;
        while (prev.next != null && prev.next.compareTo(task) < 0)
        {
          prev = prev.next;
        }
        task.next = prev.next;
        prev.next = task;
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Inner class
  // ---------------------------------------------------------------------------

  class PeriodicTask implements Comparable
  {
    long         wakeUpTime;
    long         period;
    Task         taskItf;
    PeriodicTask next;

    PeriodicTask(long period, Task taskItf)
    {
      this.period = period;
      this.taskItf = taskItf;
      wakeUpTime = System.currentTimeMillis() + period;
    }

    /**
     * @see Comparable#compareTo(Object)
     */
    public int compareTo(Object o)
    {
      PeriodicTask periodicTask = (PeriodicTask) o;
      return (int) (wakeUpTime - periodicTask.wakeUpTime);
    }
  }
}
