/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.control.activity.task.thread;

import org.objectweb.dream.control.activity.manager.EndOfThread;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Abstract class representing a thread task.
 */
public abstract class AbstractThreadTaskImpl
    implements
      Task,
      BindingController,
      LifeCycleController
{

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  /** The scheduler to which the thread is bound. */
  protected Scheduler   schedulerItf;
  protected EndOfThread endOfThreadItf;

  // ---------------------------------------------------------------------------
  // Fields
  // ---------------------------------------------------------------------------

  protected Logger      logger;
  private String        fcState;

  // ---------------------------------------------------------------------------
  // Methods that must be overriden
  // ---------------------------------------------------------------------------

  /**
   * Called by {@link #execute(Object) }method.
   * 
   * @return <code>false</code> if the {@link #execute(Object) }method must
   *         return.
   */
  protected abstract boolean isExecuting();

  /**
   * Called by {@link #execute(Object)}method. Sets its execution state.
   * 
   * @param b <code>false</code> if the {@link #execute(Object) }method must
   *          return.
   */
  protected abstract void setExecuting(boolean b);

  // ---------------------------------------------------------------------------
  // Implementation of the Task interface
  // ---------------------------------------------------------------------------

  /**
   * @see Task#execute(Object)
   */
  public Object execute(Object hints) throws InterruptedException
  {
    try
    {
      while (isExecuting())
      {
        logger.log(BasicLevel.DEBUG, "Call to schedule");
        Object o = schedulerItf.schedule(-1, null);
        if (Task.STOP_EXECUTING.equals(o))
        {
          setExecuting(endOfThreadItf.threadEnded(this));
        }
      }
    }
    catch (Throwable e)
    {
      logger.log(BasicLevel.ERROR, "Exception catched", e);
    }
    logger.log(BasicLevel.DEBUG, "End of the execute method");
    return null;
  }

  /**
   * @see Task#interrupted()
   */
  public void interrupted()
  {
    // do nothing
  }

  /**
   * @see Task#registered(Object)
   */
  public void registered(Object controlItf)
  {
    // do nothing
  }

  /**
   * @see Task#unregistered()
   */
  public void unregistered()
  {
    // do nothing
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see BindingController#bindFc(String, Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(Scheduler.ITF_NAME))
    {
      schedulerItf = (Scheduler) serverItf;
    }
    else if (clientItfName.equals(EndOfThread.ITF_NAME))
    {
      endOfThreadItf = (EndOfThread) serverItf;
    }
    else if (clientItfName.equals("logger"))
    {
      logger = (Logger) serverItf;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see BindingController#unbindFc(String)
   */
  public synchronized void unbindFc(String clientItfName)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(Scheduler.ITF_NAME))
    {
      schedulerItf = null;
    }
    else if (clientItfName.equals(EndOfThread.ITF_NAME))
    {
      endOfThreadItf = null;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see BindingController#lookupFc(String)
   */
  public Object lookupFc(String clientItfName) throws NoSuchInterfaceException
  {
    if (clientItfName.equals(Scheduler.ITF_NAME))
    {
      return schedulerItf;
    }
    else if (clientItfName.equals(EndOfThread.ITF_NAME))
    {
      return endOfThreadItf;
    }
    else
    {
      throw new NoSuchInterfaceException(clientItfName);
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Scheduler.ITF_NAME, EndOfThread.ITF_NAME};
  }

  // ---------------------------------------------------------------------------
  // Implementation of the LifeCycleController interface
  // ---------------------------------------------------------------------------

  /**
   * @see LifeCycleController#getFcState()
   */
  public String getFcState()
  {
    return fcState;
  }

  /**
   * @see LifeCycleController#stopFc()
   */
  public void stopFc() throws IllegalLifeCycleException
  {
    fcState = LifeCycleController.STOPPED;
  }

  /**
   * @see LifeCycleController#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    fcState = LifeCycleController.STARTED;
  }
}