/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.pool;

import java.lang.ref.SoftReference;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * {@link ObjectPool} implementation that use a list of {@link SoftReference} to
 * store recycle instances.
 */
public class ObjectPoolSoftReferenceImpl extends AbstractComponent
    implements
      ObjectPool
{
  private static final boolean                   DEFENSIVE_CHECKS = true;

  RecyclableSoftReference< ? extends Recyclable> first            = null;
  // private final ReferenceQueue<Recyclable> queue = new
  // ReferenceQueue<Recyclable>();
  private final Object                           lock             = new Object();

  // ---------------------------------------------------------------------------
  // Client interface
  // ---------------------------------------------------------------------------

  private RecyclableFactory                      recyclableFactoryItf;

  // ---------------------------------------------------------------------------
  // Implementation of the ObjectPool interface
  // ---------------------------------------------------------------------------

  /**
   * @see ObjectPool#newInstance()
   */
  public Recyclable newInstance()
  {
    // checkQueue();
    synchronized (lock)
    {
      while (first != null)
      {
        RecyclableSoftReference< ? extends Recyclable> ref = first;
        first = ref.nextRef;
        ref.nextRef = null;
        if (DEFENSIVE_CHECKS && ref.prevRef != null) Error.bug(logger);
        Recyclable obj = ref.get();
        if (obj != null)
        {
          return obj;
        }
      }
    }
    // pool empty, create a new instance
    Recyclable newInstance = recyclableFactoryItf.newInstance();
    newInstance.setReference(new RecyclableSoftReference<Recyclable>(
        newInstance));
    return newInstance;
  }

  /**
   * @see ObjectPool#recycleInstance(Recyclable)
   */
  @SuppressWarnings("unchecked")
  public void recycleInstance(final Recyclable recyclable)
  {
    // checkQueue();
    recyclable.recycle();
    RecyclableSoftReference ref = (RecyclableSoftReference) recyclable
        .getReference();
    synchronized (lock)
    {
      ref.nextRef = first;
      first = ref;
    }
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  // @SuppressWarnings("unchecked")
  // private void checkQueue()
  // {
  // RecyclableSoftReference ref;
  // while ((ref = (RecyclableSoftReference) queue.poll()) != null)
  // {
  // if (ref == first)
  // {
  // // ref is first
  // if (DEFENSIVE_CHECKS && ref.prevRef != null) Error.bug(logger);
  // first = ref.nextRef;
  // }
  // else
  // {
  // // ref not first
  // if (DEFENSIVE_CHECKS && ref.prevRef == null) Error.bug(null);
  // ref.prevRef.nextRef = ref.nextRef;
  // if (ref.nextRef != null)
  // {
  // ref.nextRef.prevRef = ref.prevRef;
  // }
  // }
  // ref.nextRef = null;
  // ref.prevRef = null;
  // }
  // }

  // ---------------------------------------------------------------------------
  // Inner class
  // ---------------------------------------------------------------------------

  private static final class RecyclableSoftReference<T extends Recyclable>
      extends
        SoftReference<T>
  {
    RecyclableSoftReference<T> prevRef;
    RecyclableSoftReference<T> nextRef;

    RecyclableSoftReference(T referent)
    {
      super(referent);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{RecyclableFactory.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  @Override
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(RecyclableFactory.ITF_NAME))
    {
      recyclableFactoryItf = (RecyclableFactory) serverItf;
    }
  }
}
