/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.util.Iterator;

import junit.framework.TestCase;

public class TestMessageImpl extends TestCase
{
  private static final String              NAME1 = "toto";
  private static final String              NAME2 = "titi";

  private MessageManager                   messageManager;

  private Message                 message1;
  private Message                 message2;

  private Message                 subMessage1;
  private Message                 subMessage2;

  private ChunkFactoryReference<ChunkTest> chunkFactory;
  private ChunkTest                        chunk1;
  private ChunkTest                        chunk2;

  protected void setUp() throws Exception
  {
    MessageManagerImpl mm = new MessageManagerImpl();
    mm.setDoCheck(true);
    messageManager = mm;

    message1 = messageManager.createMessage();
    message2 = messageManager.createMessage();
    subMessage1 = messageManager.createMessage();
    subMessage2 = messageManager.createMessage();

    chunkFactory = messageManager.getChunkFactory(ChunkTest.class);
    chunk1 = messageManager.createChunk(chunkFactory);
    chunk2 = messageManager.createChunk(chunkFactory);
  }

  public void testIsEmpty()
  {
    assertTrue(messageManager.isEmpty(message1));
    messageManager.addChunk(message1, NAME1, chunk1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeChunk(message1, NAME1);
    assertTrue(messageManager.isEmpty(message1));
    messageManager.addSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertTrue(messageManager.isEmpty(message1));
  }

  public void testGetChunk()
  {
    assertNull(messageManager.getChunk(message1, NAME1));
    messageManager.addChunk(message1, NAME1, chunk1);
    assertSame(chunk1, messageManager.getChunk(message1, NAME1));
    messageManager.removeChunk(message1, NAME1);
    assertNull(messageManager.getChunk(message1, NAME1));
  }

  public void testGetChunkNull()
  {
    try
    {
      messageManager.getChunk(message1, null);
      fail("getChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {
      // ok
    }
  }

  public void testAddChunk()
  {
    messageManager.addChunk(message1, NAME1, chunk1);
    assertSame(chunk1, messageManager.getChunk(message1, NAME1));
    try
    {
      messageManager.addChunk(message1, NAME1, chunk2);
      fail("addChunk should throw an exception here");
    }
    catch (ChunkAlreadyExistException e)
    {}
    messageManager.addChunk(message1, NAME2, chunk2);
    assertSame(chunk1, messageManager.getChunk(message1, NAME1));
    assertSame(chunk2, messageManager.getChunk(message1, NAME2));
  }

  public void testAddChunkNull()
  {
    try
    {
      messageManager.addChunk(null, NAME1, chunk1);
      fail("addChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.addChunk(message1, null, chunk1);
      fail("addChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.addChunk(message1, NAME1, null);
      fail("addChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
  }

  public void testAddOrReplaceChunk()
  {
    assertNull(messageManager.addOrReplaceChunk(message1, NAME1, chunk1));
    assertSame(chunk1, messageManager.getChunk(message1, NAME1));
    assertSame(chunk1, messageManager
        .addOrReplaceChunk(message1, NAME1, chunk2));
    assertSame(chunk2, messageManager.getChunk(message1, NAME1));
    assertNull(messageManager.addOrReplaceChunk(message1, NAME2, chunk1));
    assertSame(chunk2, messageManager.getChunk(message1, NAME1));
    assertSame(chunk1, messageManager.getChunk(message1, NAME2));
  }

  public void testAddOrReplaceChunkNull()
  {
    try
    {
      messageManager.addOrReplaceChunk(null, NAME1, chunk1);
      fail("addOrReplaceChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.addOrReplaceChunk(message1, null, chunk1);
      fail("addOrReplaceChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.addOrReplaceChunk(message1, NAME1, null);
      fail("addOrReplaceChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
  }

  public void testRemoveChunk1()
  {
    messageManager.addChunk(message1, NAME1, chunk1);
    assertSame(chunk1, messageManager.removeChunk(message1, NAME1));
    assertNull(messageManager.getChunk(message1, NAME1));
    try
    {
      messageManager.removeChunk(message1, NAME1);
      fail("removeChunk should throw an exception here");
    }
    catch (UnknownChunkException e)
    {}
  }

  public void testRemoveChunk2()
  {
    messageManager.addChunk(message1, NAME1, chunk1);
    messageManager.addChunk(message1, NAME2, chunk2);

    assertSame(chunk2, messageManager.removeChunk(message1, NAME2));
    assertSame(chunk1, messageManager.getChunk(message1, NAME1));

    assertNull(messageManager.getChunk(message1, NAME2));
    try
    {
      messageManager.removeChunk(message1, NAME2);
      fail("removeChunk should throw an exception here");
    }
    catch (UnknownChunkException e)
    {}
    assertSame(chunk1, messageManager.removeChunk(message1, NAME1));
    assertNull(messageManager.getChunk(message1, NAME1));
    try
    {
      messageManager.removeChunk(message1, NAME1);
      fail("removeChunk should throw an exception here");
    }
    catch (UnknownChunkException e)
    {}
  }

  public void testRemoveChunkNull()
  {
    try
    {
      messageManager.removeChunk(null, NAME1);
      fail("removeChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.removeChunk(message1, null);
      fail("removeChunk should throw an exception here");
    }
    catch (NullPointerException e)
    {}
  }

  public void testGetChunkNamesIterator1()
  {
    assertFalse(messageManager.getChunkNameIterator(message1).hasNext());

    messageManager.addChunk(message1, NAME1, chunk1);
    Iterator<String> iterator = messageManager.getChunkNameIterator(message1);
    assertTrue(iterator.hasNext());
    assertSame(NAME1, iterator.next());
    assertFalse(iterator.hasNext());

    messageManager.removeChunk(message1, NAME1);
    assertFalse(messageManager.getChunkNameIterator(message1).hasNext());
  }

  public void testGetChunkNamesIterator2()
  {
    message1.addChunk(NAME1, chunk1);
    message1.addChunk(NAME2, chunk2);
    Iterator<String> iterator = messageManager.getChunkNameIterator(message1);
    assertTrue(iterator.hasNext());
    String n1 = iterator.next();
    assertTrue(iterator.hasNext());
    if (n1 == NAME1)
    {
      assertSame(NAME2, iterator.next());
    }
    else if (n1 == NAME2)
    {
      assertSame(NAME1, iterator.next());
    }
    else
    {
      fail("Name is neither NAME1 nor NAME2");
    }
    assertFalse(iterator.hasNext());
  }

  public void testAddSubMessage1()
  {
    messageManager.addSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
  }

  public void testAddSubMessage2()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage2);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage2);
    assertTrue(messageManager.isEmpty(message1));
  }

  public void testAddSubMessage3()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertTrue(messageManager.isEmpty(message1));
  }

  public void testAddSubMessage5()
  {
    try
    {
      messageManager.addSubMessage(message1, message1);
      fail("addSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}

    messageManager.addSubMessage(message1, subMessage1);
    try
    {
      messageManager.removeSubMessage(subMessage1, message1);
      fail("addSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}
  }

  public void testAddSubMessageNull()
  {
    try
    {
      messageManager.addSubMessage(message1, null);
      fail("addSubMessage should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.addSubMessage(null, subMessage1);
      fail("addSubMessage should throw an exception here");
    }
    catch (NullPointerException e)
    {}
  }

  public void testRemoveSubMessage1()
  {
    messageManager.addSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    try
    {
      messageManager.removeSubMessage(message1, subMessage1);
      fail("removeSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}
  }

  public void testRemoveSubMessage2()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage2);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    try
    {
      messageManager.removeSubMessage(message1, subMessage1);
      fail("removeSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}
    messageManager.removeSubMessage(message1, subMessage2);
    try
    {
      messageManager.removeSubMessage(message1, subMessage2);
      fail("removeSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}
  }

  public void testRemoveSubMessage3()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertFalse(messageManager.isEmpty(message1));
    messageManager.removeSubMessage(message1, subMessage1);
    assertTrue(messageManager.isEmpty(message1));
    try
    {
      messageManager.removeSubMessage(message1, subMessage1);
      fail("removeSubMessage should throw an exception here");
    }
    catch (IllegalArgumentException e)
    {}
  }

  public void testRemoveSubMessageNull()
  {
    try
    {
      messageManager.removeSubMessage(null, subMessage1);
      fail("removeSubMessage should throw an exception here");
    }
    catch (NullPointerException e)
    {}
    try
    {
      messageManager.removeSubMessage(message1, null);
      fail("removeSubMessage should throw an exception here");
    }
    catch (NullPointerException e)
    {}
  }

  public void testGetSubMessagesIterator1()
  {
    assertFalse(messageManager.getSubMessageIterator(message1).hasNext());
    messageManager.addSubMessage(message1, subMessage1);

    Iterator<Message> iterator = messageManager
        .getSubMessageIterator(message1);
    assertTrue(iterator.hasNext());
    assertSame(subMessage1, iterator.next());
    assertFalse(iterator.hasNext());
  }

  public void testGetSubMessagesIterator2()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage2);

    Iterator<Message> iterator = messageManager
        .getSubMessageIterator(message1);
    assertTrue(iterator.hasNext());
    assertSame(subMessage1, iterator.next());
    assertTrue(iterator.hasNext());
    assertSame(subMessage2, iterator.next());
    assertFalse(iterator.hasNext());
  }

  public void testGetSubMessagesIterator3()
  {
    messageManager.addSubMessage(message1, subMessage1);
    messageManager.addSubMessage(message1, subMessage1);

    Iterator<Message> iterator = messageManager
        .getSubMessageIterator(message1);
    assertTrue(iterator.hasNext());
    assertSame(subMessage1, iterator.next());
    assertTrue(iterator.hasNext());
    assertSame(subMessage1, iterator.next());
    assertFalse(iterator.hasNext());
  }

  public void testEquals1()
  {
    UtilTest.assertMessageEquals(message1, message2, messageManager);

    messageManager.addChunk(message1, NAME1, chunk1);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);

    messageManager.addChunk(message2, NAME1, chunk1);
    UtilTest.assertMessageEquals(message1, message2, messageManager);

    messageManager.addChunk(message2, NAME2, chunk2);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);
  }

  public void testEquals2()
  {
    messageManager.addChunk(message1, NAME1, chunk1);
    messageManager.addChunk(message2, NAME1, chunk2);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);
  }

  public void testEquals3()
  {
    messageManager.addChunk(message1, NAME1, chunk1);
    messageManager.addChunk(message2, NAME2, chunk1);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);
  }

  public void testEquals4()
  {
    messageManager.addSubMessage(message1, subMessage1);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);
    messageManager.addSubMessage(message2, subMessage2);
    UtilTest.assertMessageEquals(message1, message2, messageManager);

    messageManager.addChunk(subMessage1, NAME1, chunk1);
    UtilTest.assertMessageNotEquals(message1, message2, messageManager);

    messageManager.addChunk(subMessage2, NAME1, chunk1);
    subMessage2.addChunk(NAME1, chunk1);
    UtilTest.assertMessageEquals(message1, message2, messageManager);
  }

}
